# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import OrderedDict
from pprintpp import pprint as pp
import copy
import json


class Row(object):
    """ref http://stackoverflow.com/questions/4014621/a-python-class-that-acts-like-dict"
    """
    kc_map = {}
    #ex: {'_id':{ id: 1, name: 2, ... }}
    inv_kc_map = {}
    #ex: {'id''{ 1: id , 2: name, ... }}
    __ignore_key__ = ['row_cells', 'wks', '_id', 'idx']

    def __init__(self, wks, row_cells):
        """
        input: row_cell_list
            [
                <Cell R1C1 'Timestamp'>,
                <Cell R1C2 u'\u8702\u871c\u68d2\u817f'>,
                ...
            ]
        return:
            {
                "k1": "v1",
                "k2": "v2",..
            }
        """
        self._id = wks.__dict__['_id']
        if self._id not in Row.kc_map:
            #pp(wks.__dict__['_id'])
            #Row.kc_map[self._id] = self._id
            raise Exception('Need define key_col_map')
            # TO-DO make sure how it's works
        if self._id not in Row.inv_kc_map:
            Row.inv_kc_map[self._id] = {v: k for k, v in Row.kc_map[self._id].items()}

        self.wks = wks
        self.row_cells = copy.deepcopy(row_cells)
        self.idx = row_cells[0].row
        #for cell in self.row:
        for cell in self.row_cells:
            if cell.col in Row.inv_kc_map[self._id]:
                self.__dict__[Row.inv_kc_map[self._id][cell.col]] = cell.value
            else:
                pass
                """"maybe raise Exception but it would cause adding a non-mapping cell to spread not
                avaliable."""

    def iteritems(self):
        for k, v in Row.kc_map[self._id].iteritems():
            if k == "":
                continue
            if k not in self.__dict__:
                continue
            yield (k, self.__dict__[k])

    def items(self):
        for k, v in Row.kc_map[self._id].iteritems():
            if k == "":
                continue
            if k not in self.__dict__:
                continue
            val = copy.deepcopy(self.__dict__[k])
            yield (k, val)

    def __setitem__(self, key, item):
        if key == '':
            raise Exception('setitem', 'type error')

        self.__dict__[key] = item
        if key in Row.kc_map[self._id]:
            col_idx = Row.kc_map[self._id][key]
            for cell in self.row_cells:
                if cell.col  == col_idx:
                    cell.value = item
                    self.wks.update_cells([cell])
                    break

    def __contains__(self, key):
        if key in self.__dict__:
            return key in self.__dict__

    def __getitem__(self, key):
        if key not in self.__dict__:
            print("%s is not in row" % key.encode('utf-8'))
        return self.__dict__[key]

    def __repr__(self):
        _ = copy.deepcopy(self.__dict__)
        _.pop('row_cells', None)
        _.pop('wks', None)
        _.pop('_id', None)
        #return json.dumps(_).encode('utf-8')
        #return repr(self.__dict__)
        return repr(_)

    def clear(self):
        update_cells = []
        for cell in self.row_cells:
            if cell.col in Row.inv_kc_map[self._id]:
                self.__dict__[Row.inv_kc_map[self._id][cell.col]] = ""
            cell.value = ""
            update_cells.append(cell)
        self.wks.update_cells(update_cells)

    def update(self, data_dict):
        # data as dict
        """
        input: {
            k1: v1,
            k2, v2
        }
        """
        update_cells = []
        for k, v in data_dict.items():
            if k in Row.kc_map[self._id]:
                # check if key in kc_map
                col_idx = Row.kc_map[self._id][k]
                for cell in self.row_cells:
                    if cell.col  == col_idx:
                        cell.value = v
                        update_cells.append(cell)
                        self.__dict__[k] = v
                        break
        self.wks.update_cells(update_cells)

    def kv(self):
        _ = {}
        for k, v in self.__dict__.items():
            if k in Row.__ignore_key__:
                continue
            _[k] = v
        return _

    def __eq__(self, other):
        if isinstance(other, self.__class__) or type(other) is dict:
            for k, v in self.__dict__.items():
                if k not in Row.__ignore_key__:
                    if k not in other or other[k] != self.__dict__[k]:
                        return False
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)




# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import OrderedDict
from pprintpp import pprint as pp
import copy
import json
from .row import Row

class DictSheet (object):
    def gen_alphabet_list():
        alphabet_list = []
        less26 = " ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        for s in less26:
            alphabet_list.append(s)
        for i in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
            for j in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                alphabet_list.append("%s%s" % (i, j))
        return alphabet_list 
    alphabet_list = gen_alphabet_list()

    def __init__(self, wks, mapping=None):
        self.wks = wks
        """
        input: 
            wks: working sheet
            kc_map: k, v mapping

        sheet1 =  {
            2: {}, <- Row
            3: {},
        }
        """

        self.__len__, self._width = self._get_len_width()
        self._id = wks._id
        if mapping is None:
            mapping = self._get_kc_map()
        self.mapping = mapping

        #self._mapping =  _mapping
        self.keys = self.mapping
        #self._update_row_mapping()

    def _get_kc_map(self):
        mapping = {}
        col_list = self.wks.row_values(1)
        for idx, value in enumerate(col_list, start=1):
            if value is not "":
                mapping[value.strip()] = idx
        return mapping

    @property
    def mapping(self):
        return self._mapping

    @mapping.setter
    def mapping(self, mapping):
        """
        input: 
            wks from gspread 
            mapping: {
                'col_1': 1,
                'col_2': 2,
                ...
            } or 
            mapping: [
                'col_1', 'col_2', '', 'col_3'
            ]
        output: dict of key, col mapping 
                ex: { id: 1, name: 2, ... }
        """
        if '_mapping' not in self.__dict__:
            _mapping = {}
        else:
            _mapping = self._mapping

        if type(mapping) is list:
            for idx, val in enumerate(mapping):
                _mapping[val] = idx

        if type(mapping) is dict:
            _mapping  = mapping

        self._mapping = _mapping
        # TO-DO upadte first rows in sheet 
        
        last_col = len(self.wks.row_values(1))
        if last_col>26:
            last_col = 26
        #pp(last_col)
        #pp(DictSheet.alphabet_list[last_col])
        first_row_range = "A1:%s1" % DictSheet.alphabet_list[last_col]
        #pp(first_row_range)
        cells = self.wks.range(first_row_range)
        #cells = self.wks.range("A1:Z1")
        inv_mapping = {v: k for k, v in self._mapping.items()}
        for cell in cells:
            cell.value = ""
            if cell.col in inv_mapping:
                cell.value = inv_mapping[cell.col]
        self.wks.update_cells(cells)

        self._update_row_mapping()
        _wks_len, _wks_width = self._get_len_width()
        if _wks_width > self._width:
            self._width = _wks_width
        return _mapping

    def _get_row_cells(self, idx):
        #width = DictSheet.alphabet_list[self._width]
        range_str = 'A%s:%s%s' % (str(idx), DictSheet.alphabet_list[self._width], str(idx))
        #print range_str
        row_cells = self.wks.range(range_str)
        return row_cells

    def _update_row_mapping(self):
        """
        return {
            2: {},
            3: {}, ..
        }
        """
        rows = {}
        Row.kc_map[self._id] = self._mapping
        for idx in range(2, self.__len__ + 1):
            row_cells = self._get_row_cells(idx)
            row = Row(wks=self.wks, row_cells=row_cells)
            rows[idx] = row
            self.__dict__[idx] = row
        # TO-DO 
        # self.__dict__[0] return k,v of row 1
        return rows

    def update(self, data_dict):
        """
        input:
            {idx: {k1: v1, k2:v2, ..}
            ex:
            {
                2: { u'蜂蜜棒腿': 123, ...},
                3: { u'麻辣豬肋排': 'sss', ...}
            }
        """
        try:
            for idx, data in data_dict.items():
                row_cells = self._get_row_cells(idx)
                row = Row(wks=self.wks, row_cells=row_cells)
                row.update(data)
                self.__dict__[idx] = row 
        except Exception as e:
            return -1
        return 0 

    def append(self, dict_data):
        # TO-DO: append list of dict
        idx = self.__len__  + 1
        try:
            row_cells = self._get_row_cells(idx=idx)
            row = Row(wks=self.wks, row_cells=row_cells)
            row.update(dict_data)
            self.__len__ = idx
            self.__dict__[self.__len__] = row
        except Exception as e:
            raise RuntimeError('append fails')
        return row

    def __repr__(self):
        _ = copy.deepcopy(self.__dict__)
        # return json.dumps(_).encode('utf-8')
        # return repr(self.__dict__)
        return repr(_)

    def _get_len_width(self):
        all_values = self.wks.get_all_values()
        wks_len = len(all_values)
        try:
            wks_width = max([len(_) for _ in all_values])
        except Exception as e:
            wks_width = 0
        return wks_len, wks_width

    def __idx_convert__(self, idx):
        if idx >= 2:
            return idx
        elif idx < 0:
            idx = self.__len__ + idx + 1 
            return idx
        else:
            raise IndexError('getitem', 'idx error')

    def __getitem__(self, key):
        if type(key) is int:
            idx = self.__idx_convert__(key)
            if idx not in self.__dict__:
                row_cells = self._get_row_cells(idx=idx)
                row = Row(wks=self.wks, row_cells=row_cells)
                self.__dict__[idx] = row
                if idx > self.__len__:
                    self.__len__ = idx
            return self.__dict__[idx]
        return self.__dict__[key]

    def __setitem__(self, key, item):
        if type(key) is int:
            idx = self.__idx_convert__(key)
            if idx == 1:
                raise IndexError('setitem', 'idx shoud not be 1') 
            if idx not in self.__dict__: # Insert 
                if type(item) is dict:
                    self.append(item)
                elif type(item) is Row:
                    self.__len__ += 1
                    range_str = 'A%s:%s%s' % (str(self.__len__), DictSheet.alphabet_list[self._width], str(self.__len__))
                    row_list = self.wks.range(range_str)
                    row = Row(wks=self.wks, row_cells=row_list)
                    row.update(item.kv())
                    self.__dict__[self.__len__] = row
                else:
                    raise Exception('setitem', 'type error')
            else:  # Update 
                row = self.__dict__[idx]
                row.clear()
                if type(item) is dict:
                    # replace
                    # ds[4] = {'LINE ID': 'test1', 'Email': 'email'}
                    #row = self.__dict__[idx]
                    row.update(item)
                    self.__dict__[idx] = row
                elif type(item) is Row:
                    #row = self.__dict__[idx]
                    row.update(item.kv())
                    self.__dict__[idx] = row
                else:
                    raise Exception('setitem', 'type error')
            if idx > self.__len__:
                self.__len__ = idx
        else:
            self.__dict__[key] = item

    def get(self, key):
        match = {}
        for row in self.rows:
            if key in row:
                for idx, item in row.items():
                    match[idx] = item
        return match

    def extend(self):
        pass

    def iteritems(self):
        for k in range(2, self.__len__+1):
            if k in self.__dict__:
                yield (k, self.__dict__[k])

    def items(self):
        for k in range(2, self.__len__+1):
            if k in self.__dict__:
                yield (k, self.__dict__[k])

    def __iter__(self):
        # TO-DO: Test __iter__ func    
        # http://stackoverflow.com/questions/4019971/how-to-implement-iter-self-for-a-container-object-python
        for k in range(1, self.__len__+1):
            if k in self.__dict__:
                yield self.__dict__[k]

    def get_col_mapping(self, max_len):
        if max_len <=26:
            return " ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        if max_len <=26*26:
            col_list = [' ']
            alphabets ="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            for alphabet in alphabets:
                for alphabet2 in alphabets:
                    if len(col_list) <= max_len:
                        col_list.append("%s%s" % (alphabet, alphabet2))
                    else:
                        return col_list
        raise RuntimeError('append fails')



from setuptools import setup, find_packages
#from pip.req import parse_requirements
#https://stackoverflow.com/questions/25192794/no-module-named-pip-req

def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

with open("README.rst") as file:
    long_description = file.read()

install_reqs = parse_requirements('requirements.txt')
reqs = install_reqs
# REFERENCE:
# http://stackoverflow.com/questions/14399534/how-can-i-reference-requirements-txt-for-the-install-requires-kwarg-in-setuptool

setup(
    name = 'dictsheet',
    version = '0.0.11',
    keywords = ('dictsheet', 'spreadsheet', 'gspread'),
    description = 'Dict wrapper for google spreadsheet',
    license = 'MIT License',
    install_requires = reqs,
    data_files = ['requirements.txt', 'README.md', 'LICENSE.txt'],
    url = 'https://github.com/previa/dictsheet',
    long_description = long_description,
    author = 'Chandler Huang, Xander Li',
    author_email = 'previa@gmail.com',
    
    packages = find_packages(),
    platforms = 'any',
    classifiers = [
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
)
